#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['cvtools']

package_data = \
{'': ['*']}

setup(name='cvtools-alkasm',
      version='0.4.4',
      description='Utilities for computer vision.',
      author='Alexander Reynolds',
      author_email='ar@reynoldsalexander.com',
      url='https://github.com/alkasm/cvtools',
      packages=packages,
      package_data=package_data,
      python_requires='>=3.6',
     )
